<?php
/*
 Plugin Name:		Directory Index Guard
 Plugin URI:		https://wpfix.services/directory-index-guard/
 Description:		Directory Index Guard is a small light weight plugin that prevents Apache webservers from showing directory listings, commonly called directory indexes.  The installation is easy and you can turn off directory listings for all folders with one click.
 Version:			1.2.2
 Requires at least:	4.0
 Requires PHP:		5.2
 Tested up to:		5.7
 Author:			WPFix Services
 Author URI:		https://wpfix.services
 Text Domain:		directory-index-guard
 License:			GPL v2 or later
 License URI:		http://www.gnu.org/licenses/gpl-2.0.txt

 Copyright 2021  WebWorkBench LLC (email: support@webworkbench.net)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License, version 2, as
 published by the Free Software Foundation.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

// No direct access.
defined('ABSPATH') or die();

// Main class.
class Directory_Index_Guard {

	// wordpress option keys
	const DIG_CONFIG_OPTIONS = 'directory_index_guard_configuration';
	const DIG_CONFIG_SCAN_RESULTS = 'directory_index_guard_scan_results';

	// configuration options
	const OPTION_DONATION_VISIBLE = 'donate_visible';
	const OPTION_RATE_PLUGIN_VISIBLE = 'rate_plugin_visible';
	const OPTION_LAST_SCAN_DATE = 'last_scan_date';
	const OPTION_LAST_SCAN_STATUS = 'last_scan_status';
	const OPTION_INSTALL_DATE = 'install_date';
	const OPTION_HTACCESS_LAST_MODIFIED_DATE = 'htaccess_last_modify_date';
	const OPTION_MENU_TUTORIAL = 'menu_tutorial';
	const OPTION_DISMISS_NOTIF_DATE_NOT_PROTECTED = 'dismiss_notif_date_not_protected';
	const OPTION_DISMISS_NOTIF_DATE_HTACCESS_CHANGED = 'dismiss_notif_date_htaccess_changed';

	// Define recursive directory scan level
	const DIRECTORY_INDEX_GUARD_SCAN_LEVEL = 4;

	// Define dismiss notification duration time
	const DISMISS_NOTIF_TIME_DURATION = 14 * 24 * 60 * 60; // 14 days

	// Nonce name
	const NONCE_NAME = 'dig_nonce';

	// define status types
	const STATUS_EXPOSED = 'Exposed';
	const STATUS_SAFE = 'Safe';
	const STATUS_NOTSCANNED = 'Not Scanned';
	const STATUS_NEVER_RUN = 'Never Run';

	// Define htaccess block constans
	const DIG_COMMENT_START = '### Start Directory Index Guard ###';
	const DIG_COMMENT_URL = '# https://wpfix.services/directory-index-guard/';
	const DIG_INDEXES_OFF = 'Options -Indexes';
	const DIG_COMMENT_END = '### End Directory Index Guard ###';

	// class variables
	protected static $instance = null;
	protected $options = array();
	public $pluginVersion = 0;
	public $pluginDir = '';
	public $backupDir = '';
	public $backupHtaccessFile = '';
	public $wordpressHtaccessFile = '';

	/**
	 * Creates a new Directory Index Guard object and implements singleton
	 *
	 * @return Directory Index Guard
	 */
	static function get_instance() {
		if (false == is_a(self::$instance, 'Directory_Index_Guard')) {
			self::$instance = new Directory_Index_Guard();
		}

		return self::$instance;
	}

	// get_instance

	/**
	 * Class constructor.
	 */
	private function __construct() {

		// initialize variables
		$this->pluginVersion = $this->get_plugin_version();
		$this->pluginDir = dirname(__FILE__) . DIRECTORY_SEPARATOR;
		$this->backupDir = $this->pluginDir . 'backup' . DIRECTORY_SEPARATOR;
		$this->backupHtaccessFile = $this->backupDir . '.htaccess';
		$this->wordpressHtaccessFile = ABSPATH . '.htaccess';

		// init functions
		add_action('init', array(

			$this,
			'init'));
		add_action('admin_menu', array(

			$this,
			'add_tools_submenu'));
		add_action('admin_enqueue_scripts', array(

			$this,
			'enqueue_admin_assets'));

		// register filters
		add_filter('plugin_action_links_' . plugin_basename(__FILE__), array(

			$this,
			'init_plugin_links'));
		add_filter('admin_footer_text', array(

			$this,
			'admin_footer_text'));
		add_filter('plugin_row_meta', array(

			$this,
			'plugin_meta_links'), 10, 2);

		// notification functions
		add_action('admin_post_dig_set_rate_plugin', array(

			$this,
			'setRatePlugin'));
		add_action('admin_post_dig_donate', array(

			$this,
			'setDonation'));
		add_action('activated_plugin', array(

			$this,
			'dig_plugin_activated'));
		add_action('admin_notices', array(

			$this,
			'dig_admin_notice'));

		add_action('admin_post_dig_set_dismiss_date', array(

			$this,
			'setDismissDate'));

		// ajax function
		add_action('admin_post_dig_ajax_action', array(

			$this,
			'ajaxAction'));
	}

	/**
	 * Initialize plugin
	 */
	public function init() {

		// load options
		$this->options = get_option(self::DIG_CONFIG_OPTIONS, array());

		// set defaults for options
		$change = $this->init_config_option(self::OPTION_DONATION_VISIBLE, 'true');
		$change = $this->init_config_option(self::OPTION_RATE_PLUGIN_VISIBLE, 0) || $change;
		$change = $this->init_config_option(self::OPTION_LAST_SCAN_STATUS, self::STATUS_NEVER_RUN) || $change;
		$change = $this->init_config_option(self::OPTION_LAST_SCAN_DATE, time()) || $change;
		$change = $this->init_config_option(self::OPTION_INSTALL_DATE, time()) || $change;
		$change = $this->init_config_option(self::OPTION_HTACCESS_LAST_MODIFIED_DATE, 0) || $change;
		$change = $this->init_config_option(self::OPTION_MENU_TUTORIAL, 1) || $change;

		if ($change) {
			update_option(self::DIG_CONFIG_OPTIONS, $this->options);
		}

		// create an empty scan result if needed
		if (empty(get_option(self::DIG_CONFIG_SCAN_RESULTS))) {
			update_option(self::DIG_CONFIG_SCAN_RESULTS, array());
		}
	}

	public function init_config_option($key, $default) {
		if (isset($this->options[$key])) {
			return false;
		}

		// attempt to save existing variables in v1.1.0
		$existing = get_option('directory_index_guard_' . $key);
		if (isset($existing) && $existing != '') {
			$this->options[$key] = $existing;
			delete_option('directory_index_guard_' . $key);
			return true;
		}

		if (! isset($this->options[$key])) {
			$this->options[$key] = $default;
			return true;
		}

		return false;
	}

	public function set_config_option($key, $value) {
		$this->options[$key] = $value;
		update_option(self::DIG_CONFIG_OPTIONS, $this->options);
	}

	public function get_config_option($key) {
		$result = isset($this->options[$key]) ? $this->options[$key] : null;
		return $result;
	}

	function dig_plugin_activated() {
	}

	

	/**
	 * Add tools submenu.
	 */
	public function add_tools_submenu() {
		add_management_page(__('Directory Index Guard', 'depay-widget'), __('Directory Index Guard', 'depay-widget'), 'manage_options', 'directory-index-guard', array(

			$this,
			'render_dashboard_page'));
	}

	/**
	 * Render plugin's admin dashboard page.
	 */
	public function render_dashboard_page() {
		require ($this->pluginDir . 'admin' . DIRECTORY_SEPARATOR . 'dashboard.php');
	}

	function write_log($log) {
		if (is_array($log) || is_object($log)) {
			error_log(print_r($log, true));
		} else {
			error_log($log);
		}
	}

	/**
	 * Enqueue admin assets.
	 *
	 * @param string $hook_suffix
	 *        	Hook suffix.
	 */
	public function enqueue_admin_assets($hook_suffix) {

		// Don't proceed if not on the right page
		// if ($hook_suffix != 'tools_page_directory-index-guard')
		// return;

		// enqueue style sheets
		wp_enqueue_style('directory-index-guard', plugin_dir_url(__FILE__) . 'admin/css/style.css', array(), $this->pluginVersion, 'all');
		wp_enqueue_style('directory-index-guard-sweetalert2', plugin_dir_url(__FILE__) . 'admin/css/sweetalert2.css', array(), $this->pluginVersion);

		// enqueue java scripts
		wp_enqueue_script('directory-index-guard', plugin_dir_url(__FILE__) . 'admin/js/script.js', array(

			'jquery'), $this->pluginVersion, true);
		wp_enqueue_script('directory-index-guard-sweetalert2', plugin_dir_url(__FILE__) . 'admin/js/sweetalert2.all.js', array(

			'jquery'), $this->pluginVersion, true);

		wp_localize_script('directory-index-guard', 'directory_index_guard', array(

			'loading_icon_url' => plugin_dir_url(__FILE__) . 'admin/images/download.png',
			'ajaxurl' => admin_url('admin-post.php?action=dig_ajax_action'),
			'error_message' => __('An error has occurred.', 'directory_index_guard'),
			'scanning' => __('Scanning in progress!', 'directory_index_guard'),
			'several_minutes' => __('This may take several minutes.', 'directory_index_guard'),
			'nonce' => wp_create_nonce(self::NONCE_NAME)));

		// enqueue pointer scripts if not already deactivated
		$pointers = $this->get_pointers();
		if (current_user_can('administrator') && ! empty($pointers)) {
			wp_enqueue_style('wp-pointer');

			wp_enqueue_script('directory-index-guard-pointers', plugin_dir_url(__FILE__) . 'admin/js/directory-index-guard-pointers.js', array(

				'jquery'), $this->pluginVersion, true);
			wp_enqueue_script('wp-pointer');
			wp_localize_script('wp-pointer', 'directory_index_guard_pointers', $pointers);
		}

		// fix for plugins that include their CSS files on every page
		wp_dequeue_style('uiStyleSheet');
		wp_dequeue_style('wpcufpnAdmin');
		wp_dequeue_style('unifStyleSheet');
		wp_dequeue_style('wpcufpn_codemirror');
		wp_dequeue_style('wpcufpn_codemirrorTheme');
		wp_dequeue_style('collapse-admin-css');
		wp_dequeue_style('jquery-ui-css');
		wp_dequeue_style('tribe-common-admin');
		wp_dequeue_style('file-manager__jquery-ui-css');
		wp_dequeue_style('file-manager__jquery-ui-css-theme');
		wp_dequeue_style('wpmegmaps-jqueryui');
		wp_dequeue_style('wp-botwatch-css');
	}

	function get_pointers() {
		$pointers = array();
		
		$menuTutorialVisible = (int) $this->get_config_option(self::OPTION_MENU_TUTORIAL) == 1;
					
		// if menu tutorial is visible then add pointer.
		if (!$this->is_plugin_page() && $menuTutorialVisible) {
			$welcomeMessage = '<h3>Directory Index Guard</h3>';
			$welcomeMessage .= '<p>Thank you for using the <b style="font-weight: 650;">Directory Index Guard</b> plugin.</p>';
			$welcomeMessage .= '<p>Click <a href="' . admin_url("tools.php?page=directory-index-guard") . '">Tools -> Directory Index Guard</a> to open the settings page.</p>';
			$welcomeMessage .= '<p>Then click the <i>"Turn Protection On"</i> button to prevent access to your site\'s directory indexes.</p>';

			$pointers['welcome'] = array(
				'ajaxurl' => admin_url('admin-post.php?action=dig_ajax_action'),
				'subaction' => 'dismiss_menu_tutorial',
				'target' => '#menu-tools',
				'edge' => 'left',
				'align' => 'right',
				'width' => '420',
				'content' => $welcomeMessage);
		}
		
		return $pointers;
	}

	/**
	 * Add links to plugin's description in plugins table
	 */
	function plugin_meta_links($links, $file) {
		if ($file !== plugin_basename(__FILE__)) {
			return $links;
		}

		$home_link = '<a target="_blank" href="' . $this->build_external_link('/directory-index-guard/', 'plugins-page-description_plugin-homepage') . '" title="' . __('Plugin Homepage', 'directory-index-guard') . '">' . __('Plugin Homepage', 'directory-index-guard') . '</a>';
		$support_link = '<a target="_blank" href="https://wordpress.org/support/plugin/directory-index-guard" title="' . __('Get help', 'directory-index-guard') . '">' . __('Support', 'directory-index-guard') . '</a>';
		$rate_link = '<a target="_blank" href="https://wordpress.org/support/plugin/directory-index-guard/reviews/#new-post" title="' . __('Rate the plugin', 'directory-index-guard') . '">' . __('Rate the plugin ★★★★★', 'directory-index-guard') . '</a>';

		$links[] = $home_link;
		$links[] = $support_link;
		$links[] = $rate_link;

		return $links;
	}

	/**
	 * Add powered by text in admin footer
	 */
	function admin_footer_text($text_org) {
		if (false === $this->is_plugin_page()) {
			return $text_org;
		}
		$text = '<i><a target="_blank" href="' . $this->build_external_link('/directory-index-guard/', 'admin-page-footer_plugin-homepage') . '">Directory Index Guard</a> v' . $this->pluginVersion;
		$text .= ' by <a href="' . $this->build_external_link('/', 'admin-page-footer_wpfix-homepage') . '" title="' . __('Visit our site to get more great plugins', 'directory-index-guard') . '" target="_blank">WPFix Services</a>.';
		$text .= ' Please <a target="_blank" href="https://wordpress.org/support/plugin/directory-index-guard/reviews/#new-post" title="' . __('Rate the plugin', 'directory-index-guard') . '">' . __('Rate the plugin ★★★★★', 'directory-index-guard') . '</a>.</i> ';

		return $text;
	}

	/**
	 * add plugin settings links
	 */
	public function init_plugin_links($links) {
		if ($this->isDonationVisible() == 1) {
			array_unshift($links, '<a href="' . $this->build_external_link('directory-index-guard/donate/', 'plugins-page-links_donate', 'https://www.paypal.com/donate?hosted_button_id=QDGFWBHB2P8DG') . '"><b>Donate</b></a>');
		}
		array_unshift($links, '<a href="' . admin_url('tools.php') . '?page=directory-index-guard">Settings</a>');
		return $links;
	}

	/**
	 * Loads plugin's translated strings
	 *
	 * @return null
	 */
	function load_textdomain() {
		load_plugin_textdomain('directory-index-guard', false, $this->pluginDir . 'languages' . DIRECTORY_SEPARATOR);
	}

	// load_textdomain

	/* --------------------------------------------------------------------------------------------------- */
	/* ------------------------------------- ajax action - ----------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * main method for executing ajax calls
	 */
	public function ajaxAction() {
		if (empty($_POST['subaction'])) {
			wp_send_json_error(array('message' => __('Subaction is undefined.', 'directory-index-guard')));
		}
		
		$subaction = substr(sanitize_key($_POST['subaction']), 0, 64);

		$this->write_log($subaction);
		
		// anyone can turn off the tutorial and does not need nonce check
		if ($subaction == 'dismiss_menu_tutorial') {
			$this->set_config_option(self::OPTION_MENU_TUTORIAL, 0);
			wp_send_json_success(array('message' => __('Success!', 'directory-index-guard')));
		}
				
		// admin features below
		if (! isset( $_POST['nonce'] ) || ! wp_verify_nonce( $_POST['nonce'], self::NONCE_NAME)) {
			wp_send_json_error(array('message' => __('Invalid request', 'directory-index-guard')));
		}	
		
		if (false === current_user_can('administrator')) {
			wp_send_json_error(__('You are not allowed to perform this action.', 'directory-index-guard'));
		}	

		try {
			if ($subaction == 'scan-now') {
				$this->run_scan();
			}

			if ($subaction == 'protection-on') {
				$this->turn_protection_on();
				$this->set_config_option(self::OPTION_MENU_TUTORIAL, 0);
			}

			if ($subaction == 'protection-off') {
				$this->turn_protection_off();
			}

			if ($subaction == 'protection-update') {
				$this->reset_protection();
			}

			wp_send_json_success(array('message' => __('Success!', 'directory-index-guard')));
		} catch(Exception $e) {
			wp_send_json_error(array('message' => $e->getMessage()));
		}
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ------------------------------------- notifications ----------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * add global admin notices
	 */
	function dig_admin_notice() {
		if (! $this->isProtected() && $this->shouldShowNotification('NOT_PROTECTED')) {
			$text = '<div class="notice notice-error is-dismissible">';
			$text .= '<div class="wdig-p" style="margin-top:10px; margin-bottom:15px">';
			$text .= '<p style=" font-size: medium;"><b>Directory Index Guard</b> protection turned <i>off</i>.  Apache directory listings are exposed.</p>';
			$text .= '<a id="protection-on-button" href="/wp-admin/tools.php?page=directory-index-guard#protection-on-button" class="button protection-on-button">Turn Protection On</a>';
			$text .= '<a id="dismiss-button-a" class="wdig-dismiss-link" href="/wp-admin/admin-post.php?action=dig_set_dismiss_date&notification=NOT_PROTECTED">Dismiss</a>';
			$text .= '</div></div>';
			printf($text);
		} else if ($this->htaccess_changed() && $this->shouldShowNotification('HTACCESS_CHANGED')) {
			$text = '<div class="notice notice-warning is-dismissible">';
			$text .= '<div class="wdig-p" style="margin-top:10px; margin-bottom:15px">';
			$text .= '<p style=" font-size: medium;"><b>Directory Index Guard</b> has detected changes to the .htaccess file.  Directory listings may be exposed.</p>';
			$text .= '<a id="protection-update-button" href="/wp-admin/tools.php?page=directory-index-guard#protection-update-button" class="button protection-update-button">Update Protection</a>';
			$text .= '<a id="dismiss-button-b" class="wdig-dismiss-link" href="/wp-admin/admin-post.php?action=dig_set_dismiss_date&notification=HTACCESS_CHANGED">Dismiss</a>';
			$text .= '</div></div>';
			printf($text);
		}
	}


	/**
	 * Donate
	 */
	public function setDonation() {
		$v = isset($_GET['v']) ? sanitize_text_field($_GET['v']) : null;
		$option = 1;

		if ($v == '0') {
			$option = 0;
		} else if ($v == '-1') {
			$option = - 1;
		}

		$this->set_config_option(self::OPTION_DONATION_VISIBLE, $option);
		wp_redirect(admin_url('admin.php?page=directory-index-guard'), 302);
		die();
	}

	/**
	 * Dismiss notice date
	 */
	public function setDismissDate() {
		$notification = isset($_GET['notification']) ? intval($_GET['notification']) : null;
		if (in_array($notification, array('NOT_PROTECTED', 'HTACCESS_CHANGED'))) {
			$property = 'OPTION_DISMISS_NOTIF_DATE_' . $notification;
			$this->set_config_option(constant("self::$property"), time());
		}
		wp_redirect( admin_url( 'tools.php?page=directory-index-guard' ), 302 );
		die;
	}

	private function shouldShowNotification($notification_index) {
		$property = 'OPTION_DISMISS_NOTIF_DATE_' . $notification_index;
		$dismiss_notif_date = ( int ) $this->get_config_option(constant("self::$property"));
		if ( $dismiss_notif_date <= 0 ) return true;
		return time() > $dismiss_notif_date + self::DISMISS_NOTIF_TIME_DURATION;
	}

	/**
	 * 1 = donate button
	 * 0 = thank you
	 * -1 = not visible
	 */
	public function isDonationVisible() {
		$donation = $this->get_config_option(self::OPTION_DONATION_VISIBLE);
		if (! is_numeric($donation)) {
			// just in case there is something other than a number in the database
			$this->set_config_option(self::OPTION_DONATION_VISIBLE, 1);
			$donation = 1;
		}

		return $donation;
	}

	/**
	 * Rate Plugin
	 */
	public function setRatePlugin() {
		$v = isset($_GET['v']) ? sanitize_text_field($_GET['v']) : null;
		switch ($v) {
			case 'later':
				$time = time() + 30 * 24 * 60 * 60; // +30 days
				break;
			case 'never':
				$time = - 1;
				break;
			default:
				$time = 0;
		}

		$this->set_config_option(self::OPTION_RATE_PLUGIN_VISIBLE, $time);

		wp_redirect(admin_url('admin.php?page=directory-index-guard'), 302);
		die();
	}

	public function isRatePluginVisible() {
		$remind_date = (int) $this->get_config_option(self::OPTION_RATE_PLUGIN_VISIBLE);
		return ($remind_date > - 1 && $remind_date < time());
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ------------------------------------- filesystem scan --------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * recursive scan of the file system to identify scan targets
	 * merge any previous scan result from options
	 */
	private function scan_filesystem() {
		$filesystem = array();

		// Wordpress base URL
		$filesystem[get_site_url(null, '/')] = self::STATUS_NOTSCANNED;

		// URLs of directories under the website's base URL
		$elements = $this->scan_dir(ABSPATH);
		
		foreach ($elements as $element) {
			$site_url = get_site_url(null, str_replace('\\', '/', str_replace(ABSPATH, '/', $element)));

			$filesystem[$site_url] = self::STATUS_NOTSCANNED;
		}

		$database = get_option(self::DIG_CONFIG_SCAN_RESULTS);

		// remove any items in database that are not in filesystem
		foreach ($database as $url => $status) {
			if (! isset($filesystem[$url])) {
				unset($database[$url]);
			}
		}

		// merge on the correct status
		$result = array_merge($filesystem, $database);

		// Return URLs
		return $result;
	}

	/**
	 * do the recursive scan
	 */
	private function scan_dir($dir, $level = 1) {
		if ($level > self::DIRECTORY_INDEX_GUARD_SCAN_LEVEL)
			return array();
		$out = array();
		$elements = scandir($dir);
		foreach ($elements as $element) {
			if ($element != '.' && $element != '..' && is_dir($dir . $element) && $element != 'uploads' ) {
				$current = $dir . $element;
				array_push($out, $current);
				$out = array_merge($out, $this->scan_dir($dir . $element . DIRECTORY_SEPARATOR, $level + 1));
			}
		}
		return $out;
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ------------------------------------------ url scan ----------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * Run the actual scan against all the URLs and write the result
	 * to an option in the database
	 */
	public function run_scan() {
		$filesystem = $this->scan_filesystem();

		$exposed = 0;
		$safe = 0;
		foreach ($filesystem as $url => $status) {
			$response = $this->fetch_url($url);
			// $this->write_log('Scanning URL : ' . $url);

			if ($response['code'] >= 200 && $response['code'] <= 299 && $response['has_index_string']) {
				$filesystem[$url] = self::STATUS_EXPOSED;
				$exposed = $exposed + 1;
			} else {
				$filesystem[$url] = self::STATUS_SAFE;
				$safe = $safe + 1;
			}
		}

		update_option(self::DIG_CONFIG_SCAN_RESULTS, $filesystem);
		$this->set_config_option(self::OPTION_LAST_SCAN_DATE, time());

		if ($exposed > 0) {
			$this->set_config_option(self::OPTION_LAST_SCAN_STATUS, self::STATUS_EXPOSED);
			wp_send_json_error(array(

				'message' => $this->getLastScanExposedCount() . ' Directory Indexes Exposed'));
		} else {
			$this->set_config_option(self::OPTION_LAST_SCAN_STATUS, self::STATUS_SAFE);
			wp_send_json_success(array(

				'message' => $this->getLastScanSafeCount() . ' Directory Indexes Safe'));
		}
	}

	/**
	 * call the url to check if directory listing is exposed
	 */
	private function fetch_url($url) {
		$response = wp_remote_get($url, array(

			'sslverify' => false,
			'redirection' => 10));
		$body = wp_remote_retrieve_body($response);

		return array(

			'code' => wp_remote_retrieve_response_code($response),
			'has_index_string' => preg_match('/index\s+of/i', $body));
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ----------------------------------- scan status functions ----------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	// get the count of status in filesystem scan
	public function getScanCount($countstatus) {
		$filesystem = $this->scan_filesystem();

		$count = 0;
		foreach ($filesystem as $scanstatus) {
			if ($scanstatus == $countstatus) {
				$count = $count + 1;
			}
		}

		return $count;
	}

	public function getLastScanSafeCount() {
		return $this->getScanCount(self::STATUS_SAFE);
	}

	public function getLastScanExposedCount() {
		return $this->getScanCount(self::STATUS_EXPOSED);
	}

	public function getNotScannedCount() {
		return $this->getScanCount(self::STATUS_NOTSCANNED);
	}

	public function getDirectoriesRemovedCount() {
		$filesystem = $this->scan_filesystem();
		$database = get_option(self::DIG_CONFIG_SCAN_RESULTS);

		return (count($database) - count($filesystem));
	}

	public function getLastScanDate() {
		$time = $this->get_config_option(self::OPTION_LAST_SCAN_DATE);
		$date = date('l M dS, Y', $time) . ' at ' . date('g:i a e', $time);

		return $date;
	}

	public function getLastScanStatus() {
		return $this->get_config_option(self::OPTION_LAST_SCAN_STATUS);
	}

	public function isLastScanSuccess() {
		return ($this->getLastScanStatus() == self::STATUS_SAFE);
	}

	public function isLastScanExposed() {
		return ($this->getLastScanStatus() == self::STATUS_EXPOSED);
	}

	public function getLastScanNotificationClass() {
		$status = $this->getLastScanStatus();
		if ( $status == self::STATUS_EXPOSED || $status == self::STATUS_NEVER_RUN || ! $this->isProtected()) {
			return 'notice-error';
		}

		if ($this->getNotScannedCount() > 0 || $this->htaccess_changed()) {
			return 'notice-warning';
		}

		if ($this->getLastScanStatus() == self::STATUS_SAFE && $this->isProtected()) {
			return 'wdig-notice wdig-success';
		}
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ------------------------------------------ htaccess ----------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * Turn on protection
	 */
	public function turn_protection_on($runscan = true) {

		// if there are no errors and htaccess is not protected
		if (! $this->isError() && ! $this->isProtected()) {
			// message to write to htaccess
			$output_message = PHP_EOL . self::DIG_COMMENT_START . PHP_EOL . self::DIG_COMMENT_URL . PHP_EOL . self::DIG_INDEXES_OFF . PHP_EOL . self::DIG_COMMENT_END;

			// Load htaccess and modify in memory
			if (file_exists($this->wordpressHtaccessFile)) {
				$existing_htaccess_content = file_get_contents($this->wordpressHtaccessFile);

				$new_htaccess_content = $existing_htaccess_content . PHP_EOL . $output_message;
			} else {
				$new_htaccess_content = $output_message;
			}

			// save a backup of the htaccess if it exists
			if (isset($existing_htaccess_content)) {
				file_put_contents($this->backupDir . '.htaccess_' . date('Y-m-d-H-i-s'), $existing_htaccess_content);
			}

			// check the syntax before writing
			$syntax_check = $this->check_htaccess_syntax($new_htaccess_content);
			if ($syntax_check == true) {
				// success
				file_put_contents($this->wordpressHtaccessFile, $new_htaccess_content);
				$this->set_config_option(self::OPTION_HTACCESS_LAST_MODIFIED_DATE, time());
				if ($runscan) {
					$this->run_scan();
				}
			} else {
				wp_send_json_error(array(

					'message' => 'Syntax error trying to update .htaccess.'));
			}
		}
	}

	/**
	 * Turn off protection
	 */
	public function turn_protection_off($runscan = true) {
		if (! $this->isError() && $this->isProtected() && file_exists($this->wordpressHtaccessFile)) {
			$existing_htaccess_content = file_get_contents($this->wordpressHtaccessFile);

			$regex = '|' . preg_quote(self::DIG_COMMENT_START) . '.*?' . preg_quote(self::DIG_COMMENT_END) . '|ims';

			$new_htaccess_content = rtrim(preg_replace($regex, '', $existing_htaccess_content));

			// save a backup of the htaccess
			file_put_contents($this->backupDir . '.htaccess_' . date('Y-m-d-H-i-s'), $existing_htaccess_content);

			// check the syntax before writing
			$syntax_check = $this->check_htaccess_syntax($new_htaccess_content);
			if ($syntax_check == true) {
				// success
				file_put_contents($this->wordpressHtaccessFile, $new_htaccess_content);
				$this->set_config_option(self::OPTION_HTACCESS_LAST_MODIFIED_DATE, time());
				if ($runscan) {
					$this->run_scan();
				}
			} else {
				wp_send_json_error(array(

					'message' => 'Syntax error trying to update .htaccess.'));
			}
		}
	}

	/**
	 * if .htaccess file has been edited and protection is no longer the
	 * last setting in the file.
	 */
	public function reset_protection() {
		if (! $this->isError() && $this->isProtected() && file_exists($this->wordpressHtaccessFile)) {
			$this->turn_protection_off(false);
			$this->turn_protection_on(false);
			$this->run_scan();
		}
	}

	/**
	 * Check if htaccess syntax is ok
	 */
	private function check_htaccess_syntax($content) {

		// if there are no errors
		if (! $this->isError()) {
			// create test .htaccess file
			file_put_contents($this->backupHtaccessFile, $content);

			$indexUrl = plugins_url('/backup/', $this->backupDir);
			$response = wp_remote_get($indexUrl, array(

				'sslverify' => false,
				'redirection' => 0));
			$body = wp_remote_retrieve_body($response);
			$response_code = wp_remote_retrieve_response_code($response);
			
			$secure_htaccess = '<files *.*>' . PHP_EOL . '   order allow,deny' . PHP_EOL . '   deny from all' . PHP_EOL . '</files>';

			// create a simple .htaccess that will block listings
			file_put_contents($this->backupHtaccessFile, $secure_htaccess);

			if ($response_code != 500 && $response_code != 403 && strpos($body, 'Syntax Check Success') !== 1) {
				return true;
			}
		}

		return false;
	}

	public function isProtected() {
		$content = @file_get_contents($this->wordpressHtaccessFile);

		$regex = '|' . preg_quote(self::DIG_COMMENT_START) . '.*' . preg_quote(self::DIG_INDEXES_OFF) . '.*' . preg_quote(self::DIG_COMMENT_END) . '|ims';

		return preg_match($regex, $content);
	}

	public function htaccess_changed() {
		if($this->isDigCommentLastLine()) {
			$digTime = time();
			$this->set_config_option(self::OPTION_HTACCESS_LAST_MODIFIED_DATE, $digTime);
		} else {
			$digTime = $this->get_config_option(self::OPTION_HTACCESS_LAST_MODIFIED_DATE);
		}
		
		$htTime = filemtime($this->wordpressHtaccessFile);

		return $htTime > $digTime;
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ----------------------------------- error functions ----------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * is an error of any kind
	 */
	private function isError() {
		$htaccess_noaccess = $this->isHtaccessNoAccess();
		$htaccess_readonly = $this->isHtaccessReadOnly();
		$backup_noaccess = $this->isBackupNoAccess();
		$backup_readonly = $this->isBackupReadOnly();
		$notapache = ! $this->isApache();
		$syntaxerror = $this->isSyntaxError();

		return $htaccess_noaccess || $htaccess_readonly || $backup_noaccess || $backup_readonly || $notapache || $syntaxerror;
	}

	private function isSyntaxError() {
		return (isset($_GET['syntax-invalid']));
	}

	private function isBackupNoAccess() {
		$noaccess = (! is_writable($this->backupDir));
		return $noaccess;
	}

	private function isHtaccessNoAccess() {
		$noaccess = (! file_exists($this->wordpressHtaccessFile) && ! is_writable(ABSPATH));
		return $noaccess;
	}

	private function isBackupReadOnly() {
		$readonly = (file_exists($this->backupHtaccessFile) && ! is_writable($this->backupHtaccessFile));
		return $readonly;
	}

	private function isHtaccessReadOnly() {
		$readonly = (file_exists($this->wordpressHtaccessFile) && ! is_writable($this->wordpressHtaccessFile));
		return $readonly;
	}

	private function isApache() {
		$is_apache = preg_match('/apache/i', $_SERVER['SERVER_SOFTWARE']);
		return $is_apache;
	}

	/* --------------------------------------------------------------------------------------------------- */
	/* ----------------------------------- helper functions ---------------------------------------------- */
	/* --------------------------------------------------------------------------------------------------- */

	/**
	 * Helper function for generating UTM tagged links
	 */
	function build_external_link($page, $medium, $redirectto = null) {
		$base_url = 'https://wpfix.services';

		if ('/' != $page) {
			$page = '/' . trim($page, '/') . '/';
		}
		if ($page == '//') {
			$page = '/';
		}

		$args = array();

		if (isset($redirectto) && ! $redirectto == '') {
			$args['redirectto'] = $redirectto;
		}

		// $args['debug'] = 'true';
		$args['utm_source'] = 'plugin-v' . $this->pluginVersion;
		$args['utm_medium'] = $medium;
		$args['utm_campaign'] = 'directory-index-guard-wordpress';

		$out = $base_url . $page . '?' . http_build_query($args, '', '&amp;');

		return $out;
	}

	/**
	 * Test if we're on plugin's admin page
	 *
	 * @return bool
	 */
	function is_plugin_page() {
		$current_screen = get_current_screen();

		if ($current_screen->id === 'tools_page_directory-index-guard') {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Get plugin version from file header.
	 *
	 * @return string
	 */
	function get_plugin_version() {
		$plugin_data = get_file_data(__FILE__, array(

			'version' => 'Version'), 'plugin');

		return $plugin_data['version'];
	}

	/**
	 * Checks whether DIG_COMMENT_END is in the last line of htaccess file.
	 *
	 * @return bool
	 */
	private function isDigCommentLastLine() {
		if ( ! file_exists( $this->wordpressHtaccessFile ) ) return false;
		if ( ! is_readable( $this->wordpressHtaccessFile ) ) return false;

		$htaccessContent = file_get_contents( $this->wordpressHtaccessFile );
		if ( ! preg_match( '|' . preg_quote( self::DIG_COMMENT_END ) . '\Z|m', rtrim( $htaccessContent ) ) ) return false;

		return true;
	}

	// get_plugin_version

	/**
	 * Clean up on uninstall; no action on deactive at the moment.
	 *
	 * @return null
	 */
	static function uninstall() {

		// $this->turn_protection_off(false);
		delete_option(self::DIG_CONFIG_OPTIONS);
		delete_option(self::DIG_CONFIG_SCAN_RESULTS);
	}

	// uninstall

	/**
	 * Disabled; we use singleton pattern so magic functions need to be disabled.
	 *
	 * @return null
	 */
	public function __clone() {
	}

	/**
	 * Disabled; we use singleton pattern so magic functions need to be disabled.
	 *
	 * @return null
	 */
	public function __sleep() {
	}

	/**
	 * Disabled; we use singleton pattern so magic functions need to be disabled.
	 *
	 * @return null
	 */
	public function __wakeup() {
	}

}

// Create plugin instance and hook things up
// Only in admin; there's no frontend functionality
if (is_admin()) {
	global $directory_index_guard_plugin;
	$directory_index_guard_plugin = Directory_Index_Guard::get_instance();
	add_action('plugins_loaded', array(

		$directory_index_guard_plugin,
		'load_textdomain'));
	register_uninstall_hook(__FILE__, array(

		'Directory_Index_Guard',
		'uninstall'));
}

